# EditFlag.rb (or EditFlag.rbs if scrambled.)
#
# An extension that displays an edit warning flag popup window
# when editing a component, if it's definition has multiple instances.
#
# SketchUp Ruby Extension by Dan Rathbun, AUG 2013.
# Licensed to: Ash Scott, Coastal Concepts.
 
require('sketchup.rb')
 
module CoastalConcepts # <--<<< Ash Scott's top level namespace

  module EditFlag  # <--<<< this plugin's sub namespace

    ### Constants
    #
    unless defined?(OPT_KEY)
      VERSION  = '2.4.0'
      LOAD_TAG = Module.nesting[0].name << ':' << File.basename(__FILE__)
      OPT_KEY  = 'Plugin_' << Module.nesting[0].name.gsub('::','_')
    end

    ### Module Variables (also see 'EditFlag_loader.rb' for other user vars)
    #
    @@spy = nil

    @@flagitem = nil #unless #defined?(@@flagitem)
    @@flag_use = MF_CHECKED

    @@startup   = true #unless defined?(@@startup)
    @@show_flag = true
    @@mousemenu = false

    @@debug = false

    ### Plugin Module Proxy class ------------------------------------ ###
    #
    class << self
   
      def debug()
        @@debug
      end

      def debug=(arg)
        @@debug = arg ? true : false
      end

      def extension()
        @@extension
      end

      def mousemenu()
        @@mousemenu
      end

      def mousemenu=(arg) # internally used
        bool = arg ? true : false
        @@mousemenu = bool
      end

      def remind()
        @@remind
      end

      def remind?()
        @@remind
      end

      def show_flag()
        @@show_flag
      end

      def show_flag?()
        @@show_flag
      end

      def startup()
        @@startup
      end

      def startup?()
        @@startup
      end

      def startup=(arg) # internally used
        @@startup = arg ? true : false
      end

      def toggle()
        @@startup = false if @@startup
        @@show_flag = !@@show_flag
        @@flag_use =( @@show_flag ? MF_CHECKED : MF_UNCHECKED )
        # save the new setting:
        Sketchup.write_default(OPT_KEY,'show_flag',@@show_flag)
      end
 
    end # proxy class


    ### Application Observer ----------------------------------------- ###
    #
    class AppSpy < Sketchup::AppObserver
 
      @@parent = Module.nesting[1] # ref to outer namespace
      
      def initialize(*args)
        @startup_model_hooked = false
        return self
      end
   
      def debug()
        @@parent.debug
      end

      def expectsStartupModelNotifications()
        @startup_model_hooked = true
      end

      def startup_model_hooked?()
        @startup_model_hooked
      end

      def onNewModel(model)
        #
        if debug
          puts()
          puts("onNewModel() called\nmodel: #{model.inspect}")
          puts("show_flag = #{@@parent.show_flag}")
        end
        #
        model.add_observer(@@parent::ModSpy.new(model))
        #
      end
 
      def onOpenModel(model)
        #
        if debug
          puts()
          puts("onOpenModel() called\nmodel: #{model.inspect}")
          puts("show_flag = #{@@parent.show_flag}")
        end
        #
        model.add_observer(@@parent::ModSpy.new(model))
        #
      end
 
      def onQuit()
        # remember the last show_flag setting:
        Sketchup.write_default(@@parent::OPT_KEY,'show_flag',@@parent.show_flag)
      end
 
    end # class AppSpy
 
 
    ### Model Observer ----------------------------------------------- ###
    #
    class ModSpy < Sketchup::ModelObserver
    
      WIN =( RUBY_PLATFORM =~ /(darwin)/i ? false : true )
      
      @@parent = Module.nesting[1] # ref to outer namespace
      @@say = @@parent.module_eval('@@say')
      
      ABSPATH = @@parent::ABSPATH
      HTMLPATH = File.join(ABSPATH,"EditFlag.html")

      def debug()
        @@parent.debug
      end
   
      def initialize(model,*args)
        #
        @dialog = nil
        @prev_cam = nil
        @viewing_all = false
        #
        @editpath = nil
        @num = 0
        #
        return self
        #
      end
      
      def close_flag()
        #
        if @dialog # not false if refs a UI::WebDialog object
          #
          @dialog.close if @dialog.visible?
          @dialog = nil # next GC cycle should clean it up
          #
        end
        #
      end
 
      def count_instances()
        #
        num = 0
        #
        if @editpath.size == 1
          i = @editpath.last
        else
          revpath = @editpath.reverse
          i = revpath.shift
        end
        #
        if i.is_a?(Sketchup::ComponentInstance)
          num = i.definition.instances.size
        elsif i.is_a?(Sketchup::Group)
          num = i.entities.parent.instances.size
        end
        #
        if @editpath.size > 1
          #
          revpath.each {|i|
            if i.is_a?(Sketchup::ComponentInstance)
              num = num * i.definition.instances.size
            elsif i.is_a?(Sketchup::Group)
              num = num * i.entities.parent.instances.size
            end
          }
          #
        end
        #
        return num
        #
      end
 
      def get_instances()
        #
        if @editpath.size == 1
          i = @editpath.last
          if i.is_a?(Sketchup::ComponentInstance)
            show = i.definition.instances
          elsif i.is_a?(Sketchup::Group)
            show = i.entities.parent.instances
          end
          return show
        else
          show = []
        end
        #
        revpath = @editpath.reverse
        revpath.each {|i|
          if i.is_a?(Sketchup::ComponentInstance)
            show.concat( i.definition.instances )
          elsif i.is_a?(Sketchup::Group)
            show.concat( i.entities.parent.instances )
          end
        }
        #
        return show
        #
      end
 
      def init_flag(num)
        #
        if @editpath.last.is_a?(Sketchup::Group) &&
        !@editpath.first.is_a?(Sketchup::Group)
          warn = @@say[:warn_g].%(num)
        else
          warn = @@say[:warn_c].%(num)
        end
        @dialog = new_flag(warn)
        WIN ? @dialog.show : @dialog.show_modal
        #
        puts(warn) if debug
        # ALSO show warning flag in statusbar, while mouse is still:
        Sketchup.set_status_text(warn)
        #
      end
 
      def new_flag(warn)
        #
        dialog = UI::WebDialog.new(@@say[:caption],false,@@parent::OPT_KEY,382,77,0,0,true)
        #
        dialog.set_file( HTMLPATH )
        #
        if WIN # (Trident)
          dialog.add_action_callback('get_flag_text') {|dlg,params|
            dlg.execute_script("document.getElementById('flag').innerText='#{warn}';")
          }
        else # Mac (WebKit)
          dialog.add_action_callback('get_flag_text') {|dlg,params|
            dlg.execute_script("document.getElementById('flag').textContent='#{warn}';")
          }
        end
        #
        dialog.add_action_callback('zoom') {|dlg,params|
          if @viewing_all # zoom back to previous camera:
            view = Sketchup.active_model.active_view
            if @prev_cam
              view.camera = @prev_cam
            else
              Sketchup.send_action("viewUndo:")
            end
            @viewing_all = false
          else # zoom to view all the component instances:
            view = Sketchup.active_model.active_view
            cam  = view.camera
            e = cam.eye
            t = cam.target
            u = cam.up
            p = cam.perspective?
            f = cam.fov
            @prev_cam = Sketchup::Camera.new(e,t,u,p,f)
            view.zoom( get_instances() )
            @viewing_all = true
          end
          # add in a delay to prevent unwanted double clicks:
          UI.start_timer(1.0,false) { dlg.execute_script("ok_to_zoom = true;") }
        }
        #
        return dialog
        #
      end

      def update_flag(num)
        #
        if @editpath.last.is_a?(Sketchup::Group) &&
        !@editpath.first.is_a?(Sketchup::Group)
          warn = @@say[:warn_g].%(num)
        else
          warn = @@say[:warn_c].%(num)
        end
        #
        if WIN # (Trident)
          @dialog.execute_script("document.getElementById('flag').innerText='#{warn}';")
        else # Mac (WebKit)
          @dialog.execute_script("document.getElementById('flag').textContent='#{warn}';")
        end
        #
        puts(warn) if debug
        # ALSO show warning flag in statusbar, while mouse is still:
        Sketchup.set_status_text(warn)
        #
      end

      def onActivePathChanged(model)
        #
        @editpath = model.active_path # array if in an edit context, nil if not
        if debug
          puts()
          puts("onActivePathChanged() called\neditpath: #{@editpath.inspect}")
          puts("show_flag = #{@@parent.show_flag}")
        end
        if @editpath
          #
          @num = count_instances()
          if debug
            puts("count_instances = #{@num}")
          end
          #
          @@parent.mousemenu= true if @num > 1 # toggle appears on mouse context menu
          #
          if @@parent.show_flag
            #
            if @dialog
              update_flag(@num)
            else
              init_flag(@num) if @num > 1
            end
            #
          else # it is OFF
            #
            if @@parent.startup # check if first edit after startup
              #
              if @@parent.remind # show reminder dialog
                #
                choice = UI.messagebox( @@say[:remind], MB_YESNOCANCEL )
                #
                if choice == IDYES
                  #
                  @@parent.toggle
                  if @dialog
                    update_flag(@num)
                  else
                    init_flag(@num) if @num > 1
                  end
                  #
                elsif choice == IDNO
                  #
                  @@parent.startup= false
                  #
                end  # Otherwise user cancelled or dismissed the dialog, so do
                #      nothing ( reminders continue when edit context changes. )
              else
                #
                @@parent.startup= false
                #
              end # if remind
              #
            end # if startup
            #
          end # if showing the flag
          #
        else # not in an edit context:
          #
          @@parent.mousemenu= false  # toggle NOT on mouse context menu
          #
          if @dialog
            Sketchup.set_status_text() # clear status bar message
            close_flag() # close the flag popup
          end
          #
          @num = 0
          #
        end
        #
      end # onActivePathChanged()
      
      def onDeleteModel(model)
        if debug
          editpath = model.active_path # array if in an edit context, nil if not
          puts()
          puts("onDeleteModel() called\neditpath: #{editpath.inspect}")
          puts("show_flag = #{@@parent.show_flag}")
        end
      end
      
      def onEraseAll(model)
        #
        # This callback gets called on PC before loading a New or Saved model file.
        #
        if debug
          editpath = model.active_path # array if in an edit context, nil if not
          puts()
          puts("onEraseAll() called\neditpath: #{editpath.inspect}")
          puts("show_flag = #{@@parent.show_flag}")
        end
        #
        close_flag() if @dialog
        #
      end
  
    end # class ModSpy
 
 
    ### Run ONCE at first load --------------------------------------- ###
    #
    unless file_loaded?(LOAD_TAG)
 
      #{# Load Options from saved defaults:
      #
      flag = Sketchup.read_default(OPT_KEY,'show_flag')
      if flag.nil?
        # This is the FIRST load of THIS extension
        Sketchup.write_default(OPT_KEY,'show_flag',@@show_flag)
      else
        # flag is either true || false, so
        # restore the last session setting:
        if flag != @@show_flag
          @@show_flag = !@@show_flag
          @@flag_use =( @@show_flag ? MF_CHECKED : MF_UNCHECKED )
          # save the new setting:
          Sketchup.write_default(OPT_KEY,'show_flag',@@show_flag)
        end
      end
      #}

      #{# Menus:
      #
      @@flagitem = UI::Command.new(@@say[:menuitem]) { toggle() }
      @@flagitem.set_validation_proc() { @@flag_use }
      
      UI.menu(@@topmenu).add_item(@@flagitem)
      
      UI.add_context_menu_handler {|mmenu|
        if @@mousemenu
          mmenu.add_separator
          mmenu.add_item(@@flagitem)
        end
      } #}
 
      #{# Init Observer(s):
      #
      @@spy = AppSpy.new()
      Sketchup.add_observer(@@spy)
      #
      # For AppObserver class bug in older SketchUp versions,
      # must attach the FIRST template model opened upon startup:
      #
      @tid = UI.start_timer(1.0,true) {
        #
        if @@spy.startup_model_hooked?
          UI.stop_timer(@tid) # And do nothing.
        else
          if (model = Sketchup.active_model).is_a?(Sketchup::Model) &&
          model.valid? && model.entities &&
          model.tools.active_tool_id != 0
            UI.stop_timer(@tid)
            model.add_observer( ModSpy.new(model) )
          end
        end
        #
      } #} workaround
 
      ### Mark file as loaded in $loaded_files array:
      #
      file_loaded(LOAD_TAG)
 
      ### cleanup temp vars:
      #
      flag = nil
 
    end
 
  end # module EditFlag
 
end # module CoastalConcepts